// SPDX-License-Identifier: MIT
/**
 * Get key data from Edgelock se05x.
 * Copyright (c) 2022 Atmark Techno,Inc.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <stdint.h>
#include <limits.h>
#include <errno.h>

#include <se05x/fsl_sss_ftr.h>
#include <se05x/nxLog_sss.h> // for calling nLog() from ex_sss_main_inc.h
#include <se05x/ex_sss_boot.h>
#include <se05x/fsl_sss_se05x_apis.h>
#include <se05x/nxEnsure.h>
#include <se05x/se05x_const.h>

#include <openssl/rsa.h>
#include <openssl/ec.h>
#include <openssl/pem.h>
#include <openssl/x509.h>
#include <openssl/bn.h>
#include <openssl/err.h>

#ifndef OPENSSL_VERSION_NUMBER
#error "OPENSSL_VERSION_NUMBER not defined"
#endif
#if OPENSSL_VERSION_NUMBER < 0x10100000L ||                                    \
	OPENSSL_VERSION_NUMBER >= 0x30000000L
#error "support only openssl 1.1.x."
#endif

/* used by ex_sss_main_inc.h */
static ex_sss_boot_ctx_t gex_sss_boot_ctx;
#define EX_SSS_BOOT_PCONTEXT (&gex_sss_boot_ctx)
#define EX_SSS_BOOT_DO_ERASE 0
#define EX_SSS_BOOT_EXPOSE_ARGC_ARGV 1

#include <ex_sss_main_inc.h>

#if !defined(VERSION_STRING)
#define VERSION_STRING "for testing"
#endif
#define TYPE_STR_MAX 8 // max size of type option

/* Magic bytes for OpenSSL engine */
#define MAGIC_BYTES_SE05X_OPENSSL_ENGINE_ECC_HEAD 0x10
#define MAGIC_BYTES_SE05X_OPENSSL_ENGINE_ECC_PRIV                              \
	{                                                                      \
		0xA5, 0xA6, 0xB5, 0xB6, 0xA5, 0xA6, 0xB5, 0xB6, 0x10, 0x00     \
	}

#define MAGIC_BYTES_SE05X_OPENSSL_ENGINE_RSA                                   \
	{                                                                      \
		0xA5, 0xA6, 0xB5, 0xB6                                         \
	}

/**
 * conversion_func is for converting raw bin to the format.
 */
struct format {
	const char *type_str;
	uint32_t cipher_type;
	uint32_t key_part;
	int (*conversion_func)(const char *, sss_object_t *, const uint8_t *,
			       uint32_t);
};

static struct option options[] = { { "help", no_argument, 0, 'h' },
				   { "version", no_argument, 0, 'V' },
				   { "verbose", no_argument, 0, 'v' },
				   { "type", required_argument, 0, 't' },
				   { NULL, no_argument, NULL, 0 } };

static int save_bin(const char *path, sss_object_t *obj, const uint8_t *data,
		    uint32_t datalen);
static int convert_to_pem(const char *path, sss_object_t *obj,
			  const uint8_t *data, uint32_t datalen);
static int convert_to_rsa_refkey(const char *path, sss_object_t *obj,
				 const uint8_t *data, uint32_t datalen);
static int convert_to_ecc_refkey(const char *path, sss_object_t *obj,
				 const uint8_t *data, uint32_t datalen);

/** supported format */
static struct format type_table[] = {
	{ "raw", kSSS_CipherType_NONE, kSSS_KeyPart_NONE, save_bin },
	{ "crt", kSSS_CipherType_Binary, kSSS_KeyPart_Default, convert_to_pem },
	{ "rsa", kSSS_CipherType_RSA_CRT, kSSS_KeyPart_Pair,
	  convert_to_rsa_refkey },
	{ "ecc", kSSS_CipherType_EC_NIST_P, kSSS_KeyPart_Pair,
	  convert_to_ecc_refkey },
	{ NULL, 0, 0, NULL },
};
static int use_verbose = 0;

static void usage(FILE *stream)
{
	fputs("\
Usage: se05x_getkey [options] keyid file port\n",
	      stream);

	fputs("\
Retrieve secure object from Edgelock se05x.\n",
	      stream);

	fputs("\
  keyid            32bit hex string.\n\
  file             file path of output file.\n\
  port             device file and address for T=1 over I2C.\n\
                   must be last argument due to spec of plug and trust mw.\n",
	      stream);

	fputs("\
  options:\n\
  -t --type        format type\n\
                   specify conversion format:\n\
                     raw   retrieved data as is (der format).\n\
                     cert  x509 certificate (pem format).\n\
                     rsa   rsa reference key (pem format).\n\
                     ecc   ecc reference key (pem format).\n\
  -V --version     display version number.\n\
  -v --verbose     output a diagnostic.\n\
  -h --help        display help message.\n",
	      stream);
}

static void print_version(void)
{
	printf("se05x_getkey %s\n", VERSION_STRING);
	printf("copyright (c) 2022 Atmark Techno,Inc.\n");
}

static void print_openssl_err(const char *str_func)
{
	fprintf(stderr, "failed in openssl lib (%s). \n\n", str_func);
	ERR_print_errors_fp(stderr);
}

static struct format *find_supported_format(const char *type_str)
{
	struct format *tbl;

	for (tbl = type_table; tbl->type_str; tbl++) {
		if (!strncmp(type_str, tbl->type_str, TYPE_STR_MAX)) {
			return tbl;
		}
	}
	return NULL;
}

static struct format *find_supported_format_from_obj(sss_object_t *obj)
{
	struct format *tbl;

	for (tbl = type_table; tbl->type_str; tbl++) {
		if (tbl->cipher_type == obj->cipherType &&
		    tbl->key_part == obj->objectType) {
			return tbl;
		}
	}
	return NULL;
}

static int save_bin(const char *path, sss_object_t *obj, const uint8_t *data,
		    uint32_t datalen)
{
	(void)obj;

	int result;
	FILE *fp;
	size_t num;

	fp = fopen(path, "wb+");
	if (fp == NULL) {
		fprintf(stderr, "failed to open file. %s\n", strerror(errno));
		return kStatus_SSS_Fail;
	}
	num = fwrite(data, datalen, 1, fp);
	result = fclose(fp);
	if (num != 1) {
		fprintf(stderr, "failed to write data. %s\n", strerror(errno));
		return kStatus_SSS_Fail;
	}
	if (result) {
		fprintf(stderr, " failed to close file. %s\n", strerror(errno));
		return kStatus_SSS_Fail;
	}
	return kStatus_SSS_Success;
}

static int convert_to_pem(const char *path, sss_object_t *obj,
			  const uint8_t *data, uint32_t datalen)
{
	(void)obj;

	X509 *x509;
	FILE *fp;
	int result, result_fclose;

	x509 = d2i_X509(NULL, &data, (long)datalen);
	if (x509 == NULL) {
		print_openssl_err("d2i_X509()");
		return kStatus_SSS_Fail;
	}

	fp = fopen(path, "wb+");
	if (fp == NULL) {
		fprintf(stderr, "failed to open file. %s\n", strerror(errno));
		X509_free(x509);
		return kStatus_SSS_Fail;
	}

	result = PEM_write_X509(fp, x509);
	X509_free(x509);
	result_fclose = fclose(fp);
	if (result != 1) {
		print_openssl_err("PEM_write_X509()");
		return kStatus_SSS_Fail;
	}
	if (result_fclose) {
		fprintf(stderr, " failed to close file. %s\n", strerror(errno));
		return kStatus_SSS_Fail;
	}
	return kStatus_SSS_Success;
}

static int convert_to_rsa_refkey(const char *path, sss_object_t *obj,
				 const uint8_t *data, uint32_t datalen)
{
	FILE *fp;
	int result, result_fclose;
	EVP_PKEY *pkey;
	RSA *rsa, *newrsa;
	BIGNUM *n, *e, *d, *p, *q, *dmp1, *dmq1, *iqmp;
	uint8_t prime1 = 1;
	uint8_t prime2[4]; // keyid is 32bit
	uint8_t crt_coefficient[4] = MAGIC_BYTES_SE05X_OPENSSL_ENGINE_RSA;

	/** extract modulus and public exponent from data */
	/* returned variable data is sought to the next. */
	rsa = d2i_RSA_PUBKEY(NULL, &data, (long)datalen);
	if (rsa == NULL) {
		print_openssl_err("d2i_RSA_PUBKEY()");
		return kStatus_SSS_Fail;
	}
	n = BN_dup(RSA_get0_n(rsa));
	if (n == NULL) {
		RSA_free(rsa);
		print_openssl_err("BN_dup(n)");
		return kStatus_SSS_Fail;
	}
	e = BN_dup(RSA_get0_e(rsa));
	if (e == NULL) {
		BN_free(n);
		RSA_free(rsa);
		print_openssl_err("BN_dup(e)");
		return kStatus_SSS_Fail;
	}

	if (use_verbose) {
		printf("rsa bit length %d\n", RSA_bits(rsa));
	}

	/** prepare base structure */
	newrsa = RSA_new();
	if (newrsa == NULL) {
		BN_free(n);
		BN_free(e);
		RSA_free(rsa);
		print_openssl_err("RSA_new()");
		return kStatus_SSS_Fail;
	}

	result = RSA_generate_key_ex(newrsa, RSA_bits(rsa), e, NULL);
	if (!result) {
		BN_free(n);
		BN_free(e);
		RSA_free(rsa);
		RSA_free(newrsa);
		print_openssl_err("RSA_generate_key_ex()");
		return kStatus_SSS_Fail;
	}
	d = BN_dup(RSA_get0_d(newrsa));
	if (d == NULL) {
		BN_free(n);
		BN_free(e);
		RSA_free(rsa);
		RSA_free(newrsa);
		print_openssl_err("BN_dup(d)");
		return kStatus_SSS_Fail;
	}

	/** prepare values and overwrite components */
	result = RSA_set0_key(newrsa, n, e, d);
	if (!result) {
		BN_free(n);
		BN_free(e);
		BN_free(d);
		RSA_free(rsa);
		RSA_free(newrsa);
		print_openssl_err("RSA_generate_key_ex()");
		return kStatus_SSS_Fail;
	}
	RSA_free(rsa);

	p = BN_bin2bn(&prime1, sizeof(prime1), NULL);
	if (p == NULL) {
		RSA_free(newrsa);
		print_openssl_err("BN_bin2BN(p)");
		return kStatus_SSS_Fail;
	}

	prime2[0] = (uint8_t)((obj->keyId & 0xFF000000) >> 24);
	prime2[1] = (uint8_t)((obj->keyId & 0x00FF0000) >> 16);
	prime2[2] = (uint8_t)((obj->keyId & 0x0000FF00) >> 8);
	prime2[3] = (uint8_t)((obj->keyId & 0x000000FF) >> 0);
	q = BN_bin2bn(prime2, sizeof(prime2), NULL);
	if (q == NULL) {
		BN_free(p);
		RSA_free(newrsa);
		print_openssl_err("BN_bin2BN(q)");
		return kStatus_SSS_Fail;
	}
	result = RSA_set0_factors(newrsa, p, q);
	if (!result) {
		BN_free(p);
		BN_free(q);
		RSA_free(newrsa);
		print_openssl_err("RSA_set0_factors()");
		return kStatus_SSS_Fail;
	}

	dmp1 = BN_dup(RSA_get0_dmp1(newrsa));
	if (dmp1 == NULL) {
		RSA_free(newrsa);
		print_openssl_err("BN_dup(dmp1)");
		return kStatus_SSS_Fail;
	}
	dmq1 = BN_dup(RSA_get0_dmq1(newrsa));
	if (dmq1 == NULL) {
		RSA_free(newrsa);
		BN_free(dmp1);
		print_openssl_err("BN_dup(dmq1)");
		return kStatus_SSS_Fail;
	}
	iqmp = BN_bin2bn(crt_coefficient, sizeof(crt_coefficient), NULL);
	if (iqmp == NULL) {
		RSA_free(newrsa);
		BN_free(dmp1);
		BN_free(dmq1);
		print_openssl_err("BN_bin2BN(crt coefficient)");
		return kStatus_SSS_Fail;
	}
	result = RSA_set0_crt_params(newrsa, dmp1, dmq1, iqmp);
	if (!result) {
		BN_free(dmp1);
		BN_free(dmq1);
		BN_free(iqmp);
		RSA_free(newrsa);
		print_openssl_err("RSA_set0_crt_params()");
		return kStatus_SSS_Fail;
	}

	/** output to file */
	/** initialize digital envelope library */
	OpenSSL_add_all_algorithms();
	ERR_load_BIO_strings();
	ERR_load_crypto_strings();

	pkey = EVP_PKEY_new();
	if (pkey == NULL) {
		RSA_free(newrsa);
		EVP_cleanup();
		print_openssl_err("EVP_PKEY_new()");
		return kStatus_SSS_Fail;
	}
	result = EVP_PKEY_assign_RSA(pkey, newrsa);
	if (!result) {
		RSA_free(newrsa);
		EVP_PKEY_free(pkey);
		EVP_cleanup();
		print_openssl_err("EVP_PKEY_assign_RSA()");
		return kStatus_SSS_Fail;
	}
	fp = fopen(path, "wb+");
	if (fp == NULL) {
		EVP_PKEY_free(pkey);
		EVP_cleanup();
		fprintf(stderr, "failed to open file. %s\n", strerror(errno));
		return kStatus_SSS_Fail;
	}
	result = PEM_write_PrivateKey(fp, pkey, NULL, NULL, 0, NULL, NULL);
	EVP_PKEY_free(pkey);
	result_fclose = fclose(fp);
	if (!result) {
		EVP_cleanup();
		print_openssl_err("PEM_write_PrivateKey()");
		return kStatus_SSS_Fail;
	}
	if (result_fclose) {
		EVP_cleanup();
		fprintf(stderr, " failed to close file. %s\n", strerror(errno));
		return kStatus_SSS_Fail;
	}

	EVP_cleanup();

	return kStatus_SSS_Success;
}

static int convert_to_ecc_refkey(const char *path, sss_object_t *obj,
				 const uint8_t *data, uint32_t datalen)
{
	FILE *fp;
	int result, result_fclose;
	EVP_PKEY *pkey;
	EC_KEY *ec, *newec;
	const EC_POINT *pubkey;
	const EC_GROUP *group;
	BIGNUM *privkey;
	uint32_t privkeylen;
	uint8_t key[(OPENSSL_ECC_MAX_FIELD_BITS >> 3) + 1]; // 661bit = 83byte
	uint8_t magic[10] = MAGIC_BYTES_SE05X_OPENSSL_ENGINE_ECC_PRIV;
	uint32_t offset;

	/** extract public key from data */
	/* returned variable data is sought to the next. */
	ec = d2i_EC_PUBKEY(NULL, &data, (long)datalen);
	if (ec == NULL) {
		print_openssl_err("d2i_EC_PUBKEY");
		return kStatus_SSS_Fail;
	}
	group = EC_KEY_get0_group(ec);
	if (group == NULL) {
		EC_KEY_free(ec);
		print_openssl_err("EC_KEY_get0_group()");
		return kStatus_SSS_Fail;
	}
	pubkey = EC_KEY_get0_public_key(ec);
	if (pubkey == NULL) {
		EC_KEY_free(ec);
		print_openssl_err("EC_KEY_get0_public_key()");
		return kStatus_SSS_Fail;
	}
	if (use_verbose) {
		printf("ecc curve index %d\n", EC_GROUP_get_curve_name(group));
	}

	/** prepare base structure */
	newec = EC_KEY_new();
	if (newec == NULL) {
		EC_KEY_free(ec);
		print_openssl_err("EC_KEY_new()");
		return kStatus_SSS_Fail;
	}
	result = EC_KEY_set_group(newec, group);
	if (!result) {
		EC_KEY_free(ec);
		EC_KEY_free(newec);
		print_openssl_err("EC_KEY_new()");
		return kStatus_SSS_Fail;
	}
	result = EC_KEY_generate_key(newec);
	if (!result) {
		EC_KEY_free(ec);
		EC_KEY_free(newec);
		print_openssl_err("EC_KEY_generate_key()");
		return kStatus_SSS_Fail;
	}

	/** prepare values and overwrite components */
	result = EC_KEY_set_public_key(newec, pubkey);
	if (!result) {
		EC_KEY_free(ec);
		EC_KEY_free(newec);
		print_openssl_err("EC_KEY_set_public_key()");
		return kStatus_SSS_Fail;
	}

	privkeylen = (uint32_t)EC_GROUP_order_bits(group) >> 3;
	/* for remainder of bits to fit byte */
	if (EC_GROUP_order_bits(group) & 0x7)
		privkeylen++;
	EC_KEY_free(ec);

	if (privkeylen < (uint32_t)(sizeof(magic) + sizeof(obj->keyId))) {
		EC_KEY_free(newec);
		fprintf(stderr, "invalid private key length %d\n", privkeylen);
		return kStatus_SSS_Fail;
	}
	offset = privkeylen - (uint32_t)(sizeof(magic) + sizeof(obj->keyId));
	if (use_verbose) {
		printf("key length %d(%d) offset %d\n", privkeylen,
		       privkeylen << 3, offset);
	}

	memset(key, 0, privkeylen);
	key[0] = MAGIC_BYTES_SE05X_OPENSSL_ENGINE_ECC_HEAD;
	key[offset] = (uint8_t)((obj->keyId & 0xFF000000) >> 24);
	key[offset + 1] = (uint8_t)((obj->keyId & 0x00FF0000) >> 16);
	key[offset + 2] = (uint8_t)((obj->keyId & 0x0000FF00) >> 8);
	key[offset + 3] = (uint8_t)((obj->keyId & 0x000000FF) >> 0);
	memcpy(key + offset + sizeof(obj->keyId), magic, sizeof(magic));

	privkey = BN_bin2bn(key, (int)privkeylen, NULL);
	if (privkey == NULL) {
		EC_KEY_free(newec);
		print_openssl_err("BN_bin2BN(private key)");
		return kStatus_SSS_Fail;
	}
	result = EC_KEY_set_private_key(newec, privkey);
	BN_free(privkey);
	if (!result) {
		EC_KEY_free(newec);
		print_openssl_err("EC_KEY_set_private_key()");
		return kStatus_SSS_Fail;
	}

	/** output to file */
	/** initialize digital envelope library */
	OpenSSL_add_all_algorithms();
	ERR_load_BIO_strings();
	ERR_load_crypto_strings();

	pkey = EVP_PKEY_new();
	if (pkey == NULL) {
		EC_KEY_free(newec);
		EVP_cleanup();
		print_openssl_err("EVP_PKEY_new()");
		return kStatus_SSS_Fail;
	}
	result = EVP_PKEY_assign_EC_KEY(pkey, newec);
	if (!result) {
		EC_KEY_free(newec);
		EVP_PKEY_free(pkey);
		EVP_cleanup();
		print_openssl_err("EVP_PKEY_assign_EC_KEY()");
		return kStatus_SSS_Fail;
	}
	fp = fopen(path, "wb+");
	if (fp == NULL) {
		EVP_PKEY_free(pkey);
		EVP_cleanup();
		fprintf(stderr, "failed to open file. %s\n", strerror(errno));
		return kStatus_SSS_Fail;
	}
	result = PEM_write_PrivateKey(fp, pkey, NULL, NULL, 0, NULL, NULL);
	EVP_PKEY_free(pkey);
	result_fclose = fclose(fp);
	if (!result) {
		EVP_cleanup();
		print_openssl_err("PEM_write_PrivateKey()");
		return kStatus_SSS_Fail;
	}
	if (result_fclose) {
		EVP_cleanup();
		fprintf(stderr, " failed to close file. %s\n", strerror(errno));
		return kStatus_SSS_Fail;
	}

	EVP_cleanup();

	return kStatus_SSS_Success;
}

sss_status_t ex_sss_entry(ex_sss_boot_ctx_t *pCtx)
{
	sss_status_t status = kStatus_SSS_Fail;
	int argc = gex_sss_argc;
	char *const *argv = (void *)gex_sss_argv;
	int c;
	char *endptr;

	uint64_t val;
	uint32_t keyid;
	char path_output[PATH_MAX];
	struct format *output_format = NULL;

	sss_object_t obj;
	uint8_t data[OPENSSL_RSA_MAX_MODULUS_BITS];
	size_t datalen = sizeof(data);
	size_t tmp = datalen * 8;

	while ((c = getopt_long(argc, argv, "hVvt:", options, NULL)) != -1) {
		switch (c) {
		case 0:
			break;
		case 't':
			if (optarg) {
				output_format = find_supported_format(optarg);
				if (output_format == NULL) {
					fprintf(stderr,
						"not suppported format.\n");
					usage(stderr);
					return kStatus_SSS_Fail;
				}
			} else {
				fprintf(stderr,
					"type parameter is necessary.\n");
				usage(stderr);
				return kStatus_SSS_Fail;
			}
			break;
		case 'h':
			usage(stdout);
			return kStatus_SSS_Success;
		case 'v':
			use_verbose = 1;
			printf("set verbose mode.\n");
			break;
		case 'V':
			print_version();
			return kStatus_SSS_Success;
		default:
			fprintf(stderr, "invalid option.\n");
			usage(stderr);
			return kStatus_SSS_Fail;
		}
	}
	if (optind != (argc - 3)) {
		fprintf(stderr, "incorrect number of arguments.\n");
		usage(stderr);
		return kStatus_SSS_Fail;
	}

	val = strtoul(argv[optind], &endptr, 16);
	if (!endptr || endptr[0] != '\0') {
		fprintf(stderr, "bad argument keyid %s needs to "
			"be a non-null hex integer.\n", argv[optind]);
		usage(stderr);
		return kStatus_SSS_Fail;
	}
	if (val > UINT_MAX || val == 0) {
		fprintf(stderr, 
			"keyid %s needs to be within 32bit and not zero.\n",
			argv[optind]);
		return kStatus_SSS_Fail;
	}
	keyid = (uint32_t)val;

	strncpy(path_output, argv[optind + 1], PATH_MAX - 1);

	if (use_verbose) {
		printf("arguments\n");
		printf(" keyid        0x%X\n", keyid);
		printf(" file         %s\n", path_output);
		printf(" port         %s\n", argv[optind + 2]);
	}

	/** get secure object */
	status = sss_key_object_init(&obj, &pCtx->ks);
	ENSURE_OR_GO_EXIT(status == kStatus_SSS_Success);

	status = sss_key_object_get_handle(&obj, keyid);
	if (status != kStatus_SSS_Success) {
		fprintf(stderr, "failed to retrieve handle with keyid.\n");
		fprintf(stderr, "keyid may be invalid. please check keyid.\n");
	}
	ENSURE_OR_GO_EXIT(status == kStatus_SSS_Success);

	status = sss_key_store_get_key(&pCtx->ks, &obj, data, &datalen, &tmp);
	if (status != kStatus_SSS_Success) {
		fprintf(stderr, "failed to retrieve secure object.\n");
	}
	ENSURE_OR_GO_EXIT(status == kStatus_SSS_Success);
	if (use_verbose) {
		printf("result of sss_key_store_get_key()\n");
		printf(" length       %d\n", (uint32_t)datalen);
		printf(" sss_object_t\n");
		printf("  cipherType   %d\n", obj.cipherType);
		printf("  keyid        0x%X\n", obj.keyId);
		printf("  objectType   %d\n", obj.objectType);
	}

	/** convert secure object to specified format */
	if (output_format == NULL) {
		output_format = find_supported_format_from_obj(&obj);
		if (output_format == NULL) {
			fprintf(stderr,
				"not supported format. cipherType %s "
				"objectType %d\n",
				sss_cipher_type_sz(obj.cipherType),
				obj.objectType);
			return kStatus_SSS_Fail;
		}
	}
	if ((output_format->cipher_type != kSSS_CipherType_NONE &&
	     output_format->key_part != kSSS_KeyPart_NONE) &&
	    (output_format->cipher_type != obj.cipherType ||
	     output_format->key_part != obj.objectType)) {
		fprintf(stderr, "cannot convert to specified format.\n");
		if (use_verbose) {
			printf("cipher type %02d != %02d\n",
			       output_format->cipher_type, obj.cipherType);
			printf("object type %02d != %02d\n",
			       output_format->key_part, obj.objectType);
		}
		output_format = find_supported_format_from_obj(&obj);
		if (output_format == NULL) {
			fprintf(stderr, "target object has "
					"unsupported format.\n");
		} else {
			fprintf(stderr, "target object is %s.\n",
				output_format->type_str);
		}
		return kStatus_SSS_Fail;
	}
	if (use_verbose) {
		printf("key type is %s\n", output_format->type_str);
	}
	status = output_format->conversion_func(path_output, &obj, data,
						(uint32_t)datalen);
	if (status == kStatus_SSS_Success && use_verbose) {
		printf("succeeded.\n");
	}
	return status;
exit:
	fprintf(stdout, "failure occurred in sss api (%s)\n",
		sss_status_sz(status));
	return status;
}
